#include <stdio.h>
#include <math.h>
#include "noise.c"  // noise, noise3d, turbulence

#define MAX_PREDEFS 20
#define MAX_GRAD_POINTS 16

typedef struct Color
{ double r, g, b, a;
} Color;

typedef struct Gradient
{ double point[MAX_GRAD_POINTS];
  Color color[MAX_GRAD_POINTS];
} Gradient;

typedef struct Turbulence
{ Vector amplitude;               // 0, 0, 0
  double octaves, omega, lambda;  // 6, 0.5, 2.0
} Turbulence;

typedef struct Texture
{ double (*compute)(Turbulence, Vector);
  Turbulence t;
  Vector scale;
  Vector rotate;
} Texture;


// hypot(x + sin(x+soucet noise1), y + sin(y+soucet noise2))
double wood(Turbulence t, Vector p)
{ double o = 1, l = 1, result = 0, result2 = 0;
  int i;
  for (i = 0; i < t.octaves; i++, l *= t.lambda, o *= t.omega)
  { Vector tmp1 = {p.x * l, p.y * l, p.z * l};
    Vector tmp2 = {p.y * l + 158, p.z * l + 60, p.x * l + 218};
    result += o * noise(tmp1);
    result2 += o * noise(tmp2);
  }
  double x = sin((p.x + result) * t.amplitude.x * 2*M_PI) + p.x,
         y = sin((p.y + result2) * t.amplitude.y * 2*M_PI) + p.y;
  return sqrt(x * x + y * y);
}


// noise
double bozo(Turbulence t, Vector p)
{ return noise(p);
}


// soucet fabs(0.5 - noise)
double gran(Turbulence t, Vector p)
{ double o = 1, l = 1, result = 0;
  int i;
  for (i = 0; i < t.octaves; i++, l *= t.lambda, o *= t.omega)
  { Vector tmp = {p.x * l, p.y * l, p.z * l};
    result += o * fabs(0.5 - noise(tmp));
  }
  return result;
}


// x + soucet noise
double marb(Turbulence t, Vector p)
{ double o = 1, l = 1, result = 0;
  int i;
  for (i = 0; i < t.octaves; i++, l *= t.lambda, o *= t.omega)
  { Vector tmp = {p.x * l, p.y * l, p.z * l};
    result += o * noise(tmp);
  }
  return p.x + t.amplitude.x * result;
}


// soucet bound(0, 2 * noise - 0.5, 1)
double wrin(Turbulence t, Vector p)
{ double o = 1, l = 1, result = 0;
  int i;
  for (i = 0; i < t.octaves; i++, l *= t.lambda, o *= t.omega)
  { Vector tmp = {p.x * l, p.y * l, p.z * l};
    double n = noise(tmp)*2.0 - 0.5;
    result += o * (n<0 ? 0 : n>1 ? 1 : n);
  }
  return result / 2.0;
}


Gradient gradients[] =
{ // points                       colors
 { {0,.1,.9,1}, {{.88,.6,.4},{.88,.6,.4},{.6,.3,.2},{.6,.3,.2}} },
 { {0,.1,.1,.2,.3,.4,.4,.9,1}, {{.55,.32,.2,.1},{.55,.32,.2,.5},{.55,.35,.2,.65},{.88,.6,.4,.975},{.6,.3,.2,1},{.6,.3,.2,.5},{.6,.3,.2,.65},{.88,.6,.4,.975},{.88,.6,.4,.975},{.55,.32,.2,1}} },

 { {0,.1,.1,.9,1}, {{.175,.08,.06},{.175,.08,.06},{.175,.1,.08},{.23,.13,.09},{.175,.08,.06}} },
 { {0,.1,.1,.9,1}, {{.5,.3,.2,.1},{.5,.3,.2,.5},{.5,.3,.2,.65},{.5,.25,.15,.975},{.5,.3,.2,0}} },

 { {0,.1,.9,1}, {{.75,.65,.45},{.75,.65,.45},{.7,.55,.40},{.75,.65,.45}} },
 { {0,.1,.1,.9,1}, {{.7,.6,.4,.1},{.7,.6,.4,.5},{.7,.6,.4,.65},{.75,.7,.6,.975},{.7,.6,.4,1}} },

 { {0,.3,.6,1}, {{.3,.1,.15},{.3,.15,.15},{.5,.2,.15},{.3,.1,.15}} },
 { {0,.3,.3,.4,.4,.6,.6,1}, {{.32,.15,.17,0},{.32,.13,.15,.2},{.32,.13,.13,.35},{.52,.25,.23,.4},{.52,.25,.23,.2},{.32,.13,.13,.35},{.32,.13,.15,.5},{.32,.15,.2,0}} },

 { {0,.3,.6,1}, {{.5,.35,.1},{.5,.35,.1},{.6,.45,.1},{.5,.35,.1}} },
 { {0,.1,.1,.9,1}, {{.4,.35,.1,1},{.4,0,.1,.6},{.4,0,.1,.75},{.35,.4,.15,.9},{.4,.35,.1,1}} },

 { {0,.3,.6,1}, {{.25,.1,.1},{.25,.1,.1},{.6,.15,.1},{.25,.1,.1}} },
 { {0,.3,.5,.7,.7,1}, {{.25,.1,.1,0},{.25,.1,.1,.4},{.6,.15,.1,1},{.25,.1,.1,.4},{.6,.15,.1,.4},{.25,.1,.1,0}} },

 { {0,.1,.9,1}, {{.6,.35,.2},{.6,.35,.2},{.9,.65,.3},{.6,.35,.2}} },
 { {0,.1,1}, {{.9,.65,.3,0},{.9,.65,.3,.3},{1,1,1,1}} },

 { {0,.1,.1,.9,1}, {{.45,.3,.15},{.45,.3,.15},{.55,.3,.15},{.2,.15,.05},{.45,.3,.15}} },
 { {0,.5,1}, {{.6,.35,.2,.3},{.6,.35,.2,.8},{1,1,1,1}} },

 { {0,.153,.398,.559,.729,1}, {{.439,.310,.282},{.463,.369,.259},{.541,.369,.298},{.573,.424,.286},{.494,.396,.306},{.439,.310,.282}} },
 { {0,.153,.398,.559,.729,1}, {{.729,.502,.451,.306},{.769,.686,.592,.792},{.843,.753,.718,.396},{.780,.667,.561,.976},{.741,.659,.576,.820},{.729,.502,.451,.306}} },

 { {0,.178,.356,.525,.729,1}, {{.831,.631,.569},{.925,.831,.714},{.871,.702,.659},{.831,.631,.569},{.937,.882,.820},{.831,.631,.569}} },
 { {0,.153,.398,.559,.729,1}, {{.729,.502,.451,.306},{.769,.686,.592,.792},{.843,.753,.718,.396},{.780,.667,.561,.976},{.741,.659,.576,.820},{.729,.502,.451,.306}} },

 { {0,.3,.4,.6,1}, {{0,0,0},{1,.98,.99},{.98,.99,1},{.99,1,.98},{1,1,1}} },
 { {0,.119,.322,.449,.729,1}, {{.310,.384,.420,.631},{.322,.369,.416,.847},{.424,.369,.420,.549},{.373,.424,.518,.965},{.482,.573,.533,.741},{.310,.384,.420,.631}} },

 { {0,.45,.55,.75,1}, {{.03922,.20784,.52941},{.03922,.20784,.52941},{.22745,.23922,.42353},{.20784,.3098,.54118},{.20784,.3098,.54118}} },
 { {0,.25,.45,.85,1}, {{.30196,.35294,.48627,.6},{.30196,.35294,.48627,.6},{.30196,.35294,.48627,.475},{.17647,.24706,.39216,.715},{.17647,.24706,.39216,.75}} },

 { {0,.1,.2,.5,1}, {{1,1,1},{.21,.229,.24},{.223,.209,.22},{.24,.229,.21},{.229,.229,.229}} },
 { {0,.1,.9,1}, {{0,0,0,0},{.55,.09,.09,.8},{.329,.329,.329,.9},{0,0,0,0}} },

 { {0,.153,.398,.559,.729,1}, {{.306+.729*(1-.306),.306+.502*(1-.306),.306+.451*(1-.306)},{.792+.769*(1-.792),.792+.686*(1-.792),.792+.592*(1-.792)},{.396+.843*(1-.396),.396+.753*(1-.396),.396+.718*(1-.396)},{.976+.780*(1-.976),.976+.667*(1-.976),.976+.561*(1-.976)},{.820+.741*(1-.820),.820+.659*(1-.820),.820+.576*(1-.820)},{.306+.729*(1-.306),.306+.502*(1-.306),.306+.451*(1-.306)}} },
 { {0,.216,.241,.267,.759,.784,.810,1}, {{.812,.812,.812,.835},{.745,.843,.835,.847},{.404,.337,.337,.463},{.773,.729,.745,.622},{.914,.843,.725,.651},{.153,.133,.208,.437},{.812,.812,.812,.835},{.812,.812,.812,.835}} },

 { {0,.034,.342,.462,.632,.983,1}, {{.027,.012,.012},{.851*(1-.235),.812*(1-.235),.741*(1-.235)},{.792*(1-.839),.694*(1-.839),.690*(1-.839)},{.631*(1-.608),.506*(1-.608),.471*(1-.608)},{.851*(1-.922),.812*(1-.922),.741*(1-.922)},{.647*(1-.282),.655*(1-.282),.655*(1-.282)},{.027,.012,.012}} },
 { {0,.153,.297,.441,.763,1}, {{.831,.631,.569,.447},{.925,.831,.714,.678},{.871,.702,.659,.475},{.831,.631,.569,.918},{.937,.882,.820,.655},{.831,.631,.569,.447}} },

 { {0,.241,.284,.336,.474,.810,.836,1}, {{.973,.973,.976},{.973,.973,.976},{.6,.741,.608},{.82,.643,.537},{.886,.78,.714},{.996,.643,.537},{.973,.973,.976},{.973,.973,.976}} },
 { {0,.2,.3,1}, {{1,1,1,0},{1,1,1,.7},{1,1,1,1},{1,1,1,1}} },

 { {0,.168,.301,.398,.558,.655,.735,.823,.876,.929,1}, {{.824,.725,.584},{.514,.584,.533},{.298,.376,.318},{.263,.337,.282},{.431,.506,.451},{.529,.631,.471},{.333,.376,.318},{.298,.376,.318},{.416,.376,.318},{.416,.376,.318},{.824,.725,.584}} },
 { {0,.168,.301,.398,.558,.655,.735,.823,.876,.929,1}, {{.824,.725,.584,.6},{.514,.584,.533,.6},{.298,.376,.318,.6},{.263,.337,.282,.7},{.431,.506,.451,.6},{.529,.631,.471,.5},{.333,.376,.318,.7},{.298,.376,.318,.6},{.416,.376,.318,.5},{.416,.376,.318,.6},{.824,.725,.584,.7}} },

 { {0,.155,.328,.474,.647,.810,.922,1}, {{.686,.235,.282},{.686,.235,.282},{.494,.243,.294},{.769,.329,.373},{.769,.329,.373},{.686,.235,.282},{.792,.388,.427},{.686,.235,.282}} },
 { {0,.07,.2,.3,.7,1}, {{1,1,1,0},{1,1,1,0},{.329,.329,.329,0},{1,1,1,1},{.329,.329,.329,1},{.329,.329,.329,0}} },

 { {0,1}, {{.3,.3,.3},{.7,.7,.7}} },
 { {0,1}, {{.5,.25,.1,.85},{.65,.4,0,.65}} },

 { {0,.222,.342,.393,.709,.821,1}, {{.8,.67,.25},{.8,.67,.25},{.6,.34,.04},{.8,.67,.25},{.8,.67,.25},{.53,.29,.02},{.8,.67,.25}} },
 { {0,.120,.231,.496,.701,.829,1}, {{1,1,1,1},{.70,.41,.11,.6},{.70,.46,.11,.6},{1,1,1,1},{1,1,1,1},{.70,.46,.11,.6},{1,1,1,1}} },
};


Texture textures[] =
{ //type turb:amp   octaves omega lambda  scale        rotate
  { wood, { {.04,.04}, 3,   .5,   2   },  {.05,.05,1} },
  { wood, { {.1 ,.5 }, 5,   .5,   3.25},  {.15,.5,1},  {5,10,5} },

  { wood, { {.04,.04}, 3,   .5,   2   },  {.15,.15,1} },
  { wood, { {.1 ,.35}, 5,   .5,   3.25},  {.25,.25,1}, {5,10,5} },

  { wood, { {.04,.04}, 3,   .5,   2   },  {.05,.05,1}, {2,2,0} },
  { wood, { {.1 ,.5 }, 5,   .5,   3.25},  {.15,.5,1}, {5,10,5} },

  { wood, { {.04,.04}, 4,   1.25, 2   },  {.15,.15,1} },
  { bozo, { },                            {.025,.025,1},{5,10,5} },

  { wood, { {.04,.04}, 4,   1.25, 2   },  {.15,.15,1} },
  { wood, { {.025,.025},4,  1.3,  2   },  {.3,.2,1},   {0,0,0} },

  { wood, { {.04,.06}, 4,   .5,   2   },  {.15,.15,1}, {10,0,0} },
  { wood, { {.05,.05}, 4,   .95,  2   },  {.2,.2,1},   {30,0,0} },

  { wood, { {.05,.08}, 4,   .5,   2   },  {.15,.15,1} },
  { bozo, { },                            {.01,.01,100000} },

  { wood, { {.04,.04}, 3,   .5,   2   },  {.05,.05,1} },
  { wood, { {.04,.04}, 4,   .5,   2   },  {.033,.033,.66} },

  { gran, { {}, 6, .5, 2 },               {.5,.75,.5}, {0,0,40} },
  { gran, { {}, 6, .5, 2 },               {.5,.75,.5}, {0,0,-30} },

  { marb, { {.5},      6,   .5,   2   },  {1,5,1},     {0,0,50} },
  { gran, { {}, 6, .5, 2 },               {.25,.75,.5},{0,0,-30} },

  { gran, { {}, 6, .5, 2 },               {.25,.5,.5}, {0,0,-30} },
  { gran, { {}, 6, .5, 2 },               {.25,.5,.25},{0,0,40} },

  { marb, { {1.5},     6,   .6235,2.25},  {.75,1.33,.35},{10,5,45} },
  { gran, { {}, 6, .5, 2 },               {.195,.25,.1},{5,10,-55} },

  { gran, { {}, 6, .5, 2 },               {.375,.75,.75},{0,0,40} },
  { marb, { {.85},     6,   .5,   2   },  {.5,1,.7},   {10,0,-20} },

  { gran, { {}, 6, .5, 2 },               {.5,.75,.5}, {0,0,-30} },
  { gran, { {}, 6, .5, 2 },               {1.25,.75,1},{0,0,40} },

  { gran, { {}, 6, .5, 2 },               {.75,1.5,.5},{0,0,50} },
  { gran, { {}, 6, .5, 2 },               {.25,.5,.25},{0,0,-50} },

  { gran, { {}, 6, .5, 2 },               {.25,.175,.125},{0,0,40} },
  { gran, { {}, 6, .5, 2 },               {.5,.75,.25},{10,0,-20} },

  { marb, { {.7},      6,   .5,   2   },  {1,1,2},     {20,0,30} },
  { gran, { {}, 6, .5, 2 },               {.5,.25,.5}, {40,0,-30} },

  { marb, { {.7},      6,   .5,   2   },  {1.3,2.7,1.7},{0,0,40} },
  { gran, { {}, 6, .5, 2 },               {.25,.5,.5}, {0,0,40} },

  { gran, { {}, 6, .5, 2 },               {.075,.075,.075} },
  { wrin, { {}, 6, .5, 2 },               {.3,.3,.3} },

  { wood, { {.02,.02}, 6,   .5,   2   },  {.1,.1,.1} },
  { wood, { {.01,.01}, 6,   .5,   2   },  {.5,.5,.5} },
};



double lerp(double t, double a, double b)
{ return a + t * (b - a);
}


Color gradient_get(Gradient g, double t)
{ int i;

  t -= floor(t);
  t = 2 * (t>.5 ? 1-t : t);

  for (i = 0; t >= g.point[i+1] && g.point[i] != 1.0; i++) ;

  t -= g.point[i];
  if (g.point[i+1] - g.point[i] != 0) t /= g.point[i+1] - g.point[i];
  Color res =
  { lerp(t, g.color[i].r, g.color[i+1].r),
    lerp(t, g.color[i].g, g.color[i+1].g),
    lerp(t, g.color[i].b, g.color[i+1].b),
    lerp(t, g.color[i].a, g.color[i+1].a)
  };
  return res;
}


Color texture_pixel(int predef, double x, double y, double z)
{ Vector rot_x(Vector v, double alpha) { double s = sin(alpha / 180*M_PI), c = cos(alpha / 180*M_PI); Vector res = {v.x, c*v.y + -s*v.z, s*v.y + c*v.z}; return res; }
  Vector rot_y(Vector v, double alpha) { double s = sin(alpha / 180*M_PI), c = cos(alpha / 180*M_PI); Vector res = {c*v.x + -s*v.z, v.y, s*v.x + c*v.z}; return res; }
  Vector rot_z(Vector v, double alpha) { double s = sin(alpha / 180*M_PI), c = cos(alpha / 180*M_PI); Vector res = {c*v.x + -s*v.y, s*v.x + c*v.y, v.z}; return res; }
  Vector scale(Vector v, Vector s) { Vector res = {v.x / s.x, v.y / s.y, v.z / s.z}; return res; }
  Color blend(Color a, Color b) { Color res = {lerp(b.a, b.r, a.r), lerp(b.a, b.g, a.g), lerp(b.a, b.b, a.b), 0}; return res; }

  int p = predef*2, q = p+1;
  Vector u = {x, y, z}, v = {x, y, z};

  u = rot_x(u, -3);
  u = scale(u, textures[p].scale);
  u = rot_x(u, textures[p].rotate.x);
  u = rot_y(u, textures[p].rotate.y);
  u = rot_z(u, textures[p].rotate.z);

  v = rot_x(v, -3);
  v = scale(v, textures[q].scale);
  v = rot_x(v, textures[q].rotate.x);
  v = rot_y(v, textures[q].rotate.y);
  v = rot_z(v, textures[q].rotate.z);

  return blend(gradient_get(gradients[p], textures[p].compute(textures[p].t, u)),
               gradient_get(gradients[q], textures[q].compute(textures[q].t, v)));
}


void texture(Color* pic, int predef, int xres, int yres)
{ int i, j;
  double xscale = 1.0/xres, yscale = 1.0/yres;

  for (j = 0; j < yres; j++)
    for (i = 0; i < xres; i++)
      pic[i + j*xres] = texture_pixel(predef, i*xscale*2 - 1, 1, j*yscale*2 - 1);
}


int main(int argc, char** argv)
{ int xres, yres, predef;
  Color* pic;
  int i, j;

  printf("\nTexture generator by Jan Kadlec\nUsage: texture.exe xres yres predef\n");
  if (argc != 4) return;

  xres = atoi(argv[1]); if (xres < 1 || xres > 65536) return;
  yres = atoi(argv[2]); if (yres < 1 || yres > 65536) return;
  predef = atoi(argv[3]); if (predef < 0 || predef >= MAX_PREDEFS) return;

  pic = (Color*)malloc(xres * yres * sizeof(Color));

  texture(pic, predef, xres, yres);

  char s[20]; sprintf(s, "out%02d.ppm", predef); freopen(s, "wb", stdout);
  printf("P6\n%d %d\n255\n", xres, yres);
  for (j = 0; j < yres; j++)
  { for (i = 0; i < xres; i++)
    { putchar((unsigned char)(pic[i + j*xres].r * 255));
      putchar((unsigned char)(pic[i + j*xres].g * 255));
      putchar((unsigned char)(pic[i + j*xres].b * 255));
  } }
  return 0;
}
